/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.function.document;

import org.openvpms.archetype.rules.doc.TextDocumentHandler;
import org.openvpms.component.business.domain.im.act.DocumentAct;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.component.business.domain.im.document.Document;
import org.openvpms.component.business.service.archetype.IArchetypeService;
import org.openvpms.component.business.service.archetype.helper.NodeResolver;
import org.openvpms.component.business.service.lookup.ILookupService;
import org.openvpms.component.model.object.Reference;
import org.openvpms.component.system.common.jxpath.FunctionHelper;
import org.openvpms.component.system.common.util.PropertyState;

/**
 * Document functions.
 *
 * @author Tim Anderson
 */
public class DocumentFunctions {

    /**
     * The archetype service.
     */
    private final IArchetypeService service;

    /**
     * The lookups.
     */
    private final ILookupService lookups;

    /**
     * The text document handler.
     */
    private final TextDocumentHandler handler;

    /**
     * Constructs a {@link DocumentFunctions}.
     *
     * @param service the archetype service
     */
    public DocumentFunctions(IArchetypeService service, ILookupService lookups) {
        this.service = service;
        this.lookups = lookups;
        handler = new TextDocumentHandler(service);
    }

    /**
     * Returns the value of a text node that is backed by a document if the text is too long.
     *
     * @param object the object
     * @param node   the text node
     * @return the text. May be {@code null}
     */
    public String text(Object object, String node) {
        String result = null;
        object = FunctionHelper.unwrap(object);
        if (object instanceof IMObject) {
            NodeResolver resolver = new NodeResolver((IMObject) object, service, lookups);
            PropertyState state = resolver.resolve(node);
            if (state.getValue() == null) {
                if (state.getParent() instanceof DocumentAct) {
                    result = getText((DocumentAct) state.getParent());
                }
            } else {
                result = state.getValue().toString();
            }
        }
        return result;
    }

    /**
     * Returns the content of a document, if it is plain text.
     *
     * @param object the document act
     * @return the text. May be {@code null}
     */
    private String getText(DocumentAct object) {
        String result = null;
        Reference reference = object.getDocument();
        if (reference != null) {
            Document document = service.get(object.getDocument(), Document.class);
            if (document != null && TextDocumentHandler.TEXT_PLAIN.equals(document.getMimeType())) {
                result = handler.toString(document);
            }
        }
        return result;
    }
}
