/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.test.builder.customer.account;

import org.openvpms.archetype.test.builder.object.AbstractTestIMObjectBuilder;
import org.openvpms.component.model.act.ActRelationship;
import org.openvpms.component.model.act.FinancialAct;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.model.party.Party;
import org.openvpms.component.model.user.User;
import org.openvpms.component.service.archetype.ArchetypeService;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Set;

/**
 * Builds <em>act.customerAccountCharges*</em> instances, for testing purposes.
 *
 * @author Tim Anderson
 */
public abstract class AbstractTestCustomerChargeBuilder<B extends AbstractTestCustomerChargeBuilder<B, I>,
        I extends AbstractTestCustomerChargeItemBuilder<B, I>>
        extends AbstractTestIMObjectBuilder<FinancialAct, B> {

    /**
     * The start time.
     */
    private Date startTime;

    /**
     * The customer.
     */
    private Party customer;

    /**
     * The clinician.
     */
    private User clinician;

    /**
     * The status.
     */
    private String status;

    /**
     * The charge items.
     */
    private List<FinancialAct> items = new ArrayList<>();

    /**
     * Constructs an {@link AbstractTestCustomerChargeBuilder}.
     *
     * @param service the archetype service
     */
    public AbstractTestCustomerChargeBuilder(String archetype, ArchetypeService service) {
        super(archetype, FinancialAct.class, service);
    }

    /**
     * Sets the start time.
     *
     * @param startTime the start time
     * @return this
     */
    public B startTime(Date startTime) {
        this.startTime = startTime;
        return getThis();
    }

    /**
     * Sets the customer.
     *
     * @param customer the customer
     * @return this
     */
    public B customer(Party customer) {
        this.customer = customer;
        return getThis();
    }

    /**
     * Sets the clinician.
     *
     * @param clinician the clinician
     * @return this
     */
    public B clinician(User clinician) {
        this.clinician = clinician;
        return getThis();
    }

    /**
     * Sets the status.
     *
     * @param status the status
     * @return this
     */
    public B status(String status) {
        this.status = status;
        return getThis();
    }

    /**
     * Adds charge items.
     *
     * @param items the items to add
     * @return this
     */
    public B add(FinancialAct ... items) {
        return add(Arrays.asList(items));
    }

    /**
     * Adds charge items.
     *
     * @param items the items to add
     * @return this
     */
    public B add(List<FinancialAct> items) {
        this.items.addAll(items);
        return getThis();
    }

    /**
     * Returns a builder to add a charge item.
     *
     * @return a charge item builder
     */
    public abstract I item();

    /**
     * Builds the object.
     *
     * @param object the object
     * @param bean   a bean wrapping the object
     * @param toSave objects to save, if the object is to be saved
     */
    @Override
    protected void build(FinancialAct object, IMObjectBean bean, Set<IMObject> toSave) {
        super.build(object, bean, toSave);
        if (startTime != null) {
            bean.setValue("startTime", startTime);
        }
        if (customer != null) {
            bean.setTarget("customer", customer);
        }
        if (clinician != null) {
            bean.setTarget("clinician", clinician);
        }
        if (status != null) {
            object.setStatus(status);
        }
        BigDecimal amount = BigDecimal.ZERO;
        BigDecimal tax = BigDecimal.ZERO;
        int sequence = 0;
        for (FinancialAct item : items) {
            ActRelationship relationship = (ActRelationship) bean.addTarget("items", item);
            relationship.setSequence(sequence++);
            item.addActRelationship(relationship);
            toSave.add(item);
            IMObjectBean itemBean = bean.getBean(item);
            tax = tax.add(itemBean.getBigDecimal("tax", BigDecimal.ZERO));
            amount = amount.add(itemBean.getBigDecimal("total", BigDecimal.ZERO));
        }
        bean.setValue("amount", amount);
        bean.setValue("tax", tax);
        items.clear(); // can't reuse
    }
}
