/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.test.builder.object;

import org.openvpms.archetype.test.TestHelper;
import org.openvpms.component.model.bean.IMObjectBean;

/**
 * Value strategy.
 *
 * @author Tim Anderson
 */
public class ValueStrategy {

    /**
     * The value.
     */
    private Object value;

    /**
     * If {@code true}, the value is a prefix for a randomly generated string.
     */
    private boolean prefix;

    /**
     * Indicates that a node shouldn't be assigned a value.
     */
    private static ValueStrategy USE_DEFAULT = new ValueStrategy(null, false);

    /**
     * Constructs a {@link ValueStrategy}.
     *
     * @param value  the value
     * @param prefix if {@code true}, the value is a prefix for a randomly generated string
     */
    private ValueStrategy(Object value, boolean prefix) {
        this.value = value;
        this.prefix = prefix;
    }

    /**
     * Returns the value.
     *
     * @return the value. May be {@code null}
     */
    public Object getValue() {
        if (prefix) {
            if (value != null) {
                return TestHelper.randomName(value.toString());
            } else {
                return TestHelper.randomName("");
            }
        }
        return value;
    }

    /**
     * Returns a string version of the value.
     * <p/>
     * For random values, this returns a different result each time.
     *
     * @return a string version of the value
     */
    @Override
    public String toString() {
        Object value = getValue();
        return (value != null) ? value.toString() : null;
    }

    /**
     * Populates the named node with the value generated according to the strategy.
     * <p/>
     * The one exception is the {@link #defaultValue()} strategy, which leaves the node unchanged.
     *
     * @param bean the bean to populate
     * @param name the node name
     * @return the value. May be {@code null}
     */
    public Object setValue(IMObjectBean bean, String name) {
        Object value;
        if (!useDefault()) {
            value = getValue();
            bean.setValue(name, value);
        } else if (bean.hasNode(name)) {
            value = bean.getValue(name);
        } else {
            value = null;
        }
        return value;
    }

    /**
     * Determines if the default value should be used.
     *
     * @return {@code true} if the default value should be used
     */
    public boolean useDefault() {
        return this == USE_DEFAULT;
    }

    /**
     * Creates a strategy to indicate that the node should not be populated.
     *
     * @return the strategy
     */
    public static ValueStrategy defaultValue() {
        return USE_DEFAULT;
    }

    /**
     * Creates a strategy to populate a node with a value.
     *
     * @param value the value. May be {@code null}
     * @return the strategy
     */
    public static ValueStrategy value(Object value) {
        return new ValueStrategy(value, false);
    }

    /**
     * Creates a strategy to populate a node with a random value.
     *
     * @return the strategy
     */
    public static ValueStrategy random() {
        return random("");
    }

    /**
     * Creates a strategy to populate a node with a random value, with the specified prefix.
     *
     * @param prefix the prefix
     * @return the strategy
     */
    public static ValueStrategy random(String prefix) {
        return new ValueStrategy(prefix, true);
    }

}
