/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.test.builder.party;

import org.openvpms.archetype.rules.party.ContactArchetypes;
import org.openvpms.archetype.test.builder.entity.AbstractTestEntityBuilder;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.model.party.Contact;
import org.openvpms.component.model.party.Party;
import org.openvpms.component.service.archetype.ArchetypeService;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

/**
 * Builder for <em>party.*</em> instances, for testing purposes.
 *
 * @author Tim Anderson
 */
public abstract class AbstractTestPartyBuilder<T extends Party, B extends AbstractTestPartyBuilder<T, B>>
        extends AbstractTestEntityBuilder<T, B> {

    /**
     * The contacts.
     */
    private List<Contact> contacts = new ArrayList<>();

    /**
     * Constructs an {@link AbstractTestPartyBuilder}.
     *
     * @param archetype the archetype
     * @param type      the type
     * @param service   the archetype service
     */
    public AbstractTestPartyBuilder(String archetype, Class<T> type, ArchetypeService service) {
        super(archetype, type, service);
    }

    /**
     * Adds a contact.
     *
     * @param contact the contact to add
     * @return this
     */
    public B addContact(Contact contact) {
        contacts.add(contact);
        return getThis();
    }

    /**
     * Adds a phone contact.
     *
     * @param phone    the phone number
     * @param purposes the purposes
     * @return this
     */
    public B addPhone(String phone, String... purposes) {
        TestPhoneContactBuilder builder = new TestPhoneContactBuilder(getService());
        contacts.add(builder.phone(phone).purposes(purposes).build(false));
        return getThis();
    }

    /**
     * Adds a mobile phone contact.
     *
     * @param phone the phone number
     * @return this
     */
    public B addMobilePhone(String phone) {
        return addPhone(phone, ContactArchetypes.MOBILE_PURPOSE);
    }

    /**
     * Adds an email contact.
     *
     * @param email the email address
     * @return this
     */
    public B addEmail(String email) {
        TestEmailContactBuilder builder = new TestEmailContactBuilder(getService());
        contacts.add(builder.email(email).build(false));
        return getThis();
    }

    /**
     * Adds a location (aka address) contact.
     *
     * @param address    the street address
     * @param suburbCode the suburb code
     * @param stateCode  the state code
     * @param postCode   the post code
     * @param purposes   the contact purposes
     * @return this
     */
    public B addLocation(String address, String suburbCode, String stateCode, String postCode, String... purposes) {
        TestLocationContactBuilder builder = new TestLocationContactBuilder(getService());
        Contact contact = builder.address(address).suburbCode(suburbCode).stateCode(stateCode).postCode(postCode)
                .purposes(purposes)
                .build();
        contacts.add(contact);
        return getThis();
    }

    /**
     * Adds a website contact.
     *
     * @param url the website url
     * @return this
     */
    public B addWebsite(String url) {
        TestWebsiteContactBuilder builder = new TestWebsiteContactBuilder(getService());
        Contact contact = builder.url(url).build();
        contacts.add(contact);
        return getThis();
    }

    /**
     * Builds the party.
     *
     * @param object the party to build
     * @param bean   a bean wrapping the party
     * @param toSave objects to save, if the entity is to be saved
     */
    protected void build(T object, IMObjectBean bean, Set<IMObject> toSave) {
        super.build(object, bean, toSave);
        for (Contact contact : contacts) {
            object.addContact(contact);
        }
        contacts.clear(); // can't be reused
    }
}
