/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.test.builder.patient;

import org.openvpms.archetype.test.TestHelper;
import org.openvpms.archetype.test.builder.object.AbstractTestIMObjectBuilder;
import org.openvpms.archetype.test.builder.object.ValueStrategy;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.model.party.Party;
import org.openvpms.component.model.user.User;
import org.openvpms.component.service.archetype.ArchetypeService;

import java.util.Date;
import java.util.Set;

/**
 * Builder for patient acts, for testing purposes.
 *
 * @author Tim Anderson
 */
public abstract class AbstractTestPatientActBuilder<T extends Act, B extends AbstractTestPatientActBuilder<T, B>>
        extends AbstractTestIMObjectBuilder<T, B> {

    /**
     * The start time.
     */
    private ValueStrategy startTime = ValueStrategy.defaultValue();

    /**
     * The end time.
     */
    private ValueStrategy endTime = ValueStrategy.defaultValue();

    /**
     * The patient.
     */
    private Party patient;

    /**
     * The practice location.
     */
    private Party location;

    /**
     * The clinician.
     */
    private User clinician;

    /**
     * The act status.
     */
    private String status;

    /**
     * The act reason.
     */
    private String reason;

    /**
     * Constructs an {@link AbstractTestPatientActBuilder}.
     *
     * @param archetype the archetype to build
     * @param type      the type
     * @param service   the archetype service
     */
    public AbstractTestPatientActBuilder(String archetype, Class<T> type, ArchetypeService service) {
        super(archetype, type, service);
    }

    /**
     * Sets the start time.
     *
     * @param startTime the start time
     * @return this
     */
    public B startTime(String startTime) {
        return startTime(parseDate(startTime));
    }

    /**
     * Sets the start time.
     *
     * @param startTime the start time
     * @return this
     */
    public B startTime(Date startTime) {
        return startTime(ValueStrategy.value(startTime));
    }

    /**
     * Sets the start time.
     *
     * @param startTime the start time
     * @return this
     */
    public B startTime(ValueStrategy startTime) {
        this.startTime = startTime;
        return getThis();
    }

    /**
     * Sets the end time.
     *
     * @param endTime the end time
     * @return this
     */
    public B endTime(String endTime) {
        return endTime(parseDate(endTime));
    }

    /**
     * Sets the end time.
     *
     * @param endTime the end time
     * @return this
     */
    public B endTime(Date endTime) {
        return endTime(ValueStrategy.value(endTime));
    }

    /**
     * Sets the end time.
     *
     * @param endTime the end time
     * @return this
     */
    public B endTime(ValueStrategy endTime) {
        this.endTime = endTime;
        return getThis();
    }

    /**
     * Sets the patient.
     *
     * @param patient the patient
     * @return this
     */
    public B patient(Party patient) {
        this.patient = patient;
        return getThis();
    }

    /**
     * Sets the practice location.
     *
     * @param location the practice location
     * @return this
     */
    public B location(Party location) {
        this.location = location;
        return getThis();
    }

    /**
     * Sets the clinician.
     *
     * @param clinician the clinician
     * @return this
     */
    public B clinician(User clinician) {
        this.clinician = clinician;
        return getThis();
    }

    /**
     * Sets the status.
     *
     * @param status the status
     * @return this
     */
    public B status(String status) {
        this.status = status;
        return getThis();
    }

    /**
     * Sets the reason.
     *
     * @param reason the reason
     * @return this
     */
    public B reason(String reason) {
        this.reason = reason;
        return getThis();
    }

    /**
     * Builds the object.
     *
     * @param object the object
     * @param bean   a bean wrapping the object
     * @param toSave objects to save, if the object is to be saved
     */
    @Override
    protected void build(T object, IMObjectBean bean, Set<IMObject> toSave) {
        super.build(object, bean, toSave);

        startTime.setValue(bean, "startTime");
        endTime.setValue(bean, "endTime");

        bean.setTarget("patient", patient);
        if (location != null) {
            bean.setTarget("location", location);
        }
        if (clinician != null) {
            bean.setTarget("clinician", clinician);
        }
        if (status != null) {
            object.setStatus(status);
        }
        if (reason != null) {
            object.setReason(reason);
        }
    }

    private Date parseDate(String value) {
        if (value != null) {
            return value.contains(":") ? TestHelper.getDatetime(value) : TestHelper.getDate(value);
        }
        return null;
    }
}
