/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.archetype.test.builder.product;

import org.openvpms.archetype.rules.product.ProductPriceTestHelper;
import org.openvpms.archetype.test.builder.entity.AbstractTestEntityBuilder;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.entity.Entity;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.model.product.Product;
import org.openvpms.component.model.product.ProductPrice;
import org.openvpms.component.service.archetype.ArchetypeService;

import java.math.BigDecimal;
import java.util.Date;
import java.util.Set;

/**
 * Builder for <em>product.*</em> archetypes, for testing purposes.
 *
 * @author Tim Anderson
 */
public abstract class AbstractTestProductBuilder<T extends AbstractTestProductBuilder<T>>
        extends AbstractTestEntityBuilder<Product, T> {

    /**
     * The fixed price.
     */
    private BigDecimal fixedPrice;

    /**
     * The unit price.
     */
    private BigDecimal unitPrice;

    /**
     * The product discounts.
     */
    private Entity[] discounts;


    /**
     * The product type.
     */
    private Entity type;

    /**
     * Constructs a {@link AbstractTestProductBuilder}.
     *
     * @param archetype the product archetype
     * @param service   the archetype service
     */
    public AbstractTestProductBuilder(String archetype, ArchetypeService service) {
        super(archetype, Product.class, service);
        name("zproduct");
    }

    /**
     * Adds a fixed price.
     *
     * @param fixedPrice the fixed price
     * @return this
     */
    public T fixedPrice(BigDecimal fixedPrice) {
        this.fixedPrice = fixedPrice;
        return getThis();
    }

    /**
     * Adds a unit price.
     *
     * @param unitPrice the fixed price
     * @return this
     */
    public T unitPrice(BigDecimal unitPrice) {
        this.unitPrice = unitPrice;
        return getThis();
    }

    /**
     * Sets the product type.
     *
     * @param type the product type. An <em>entity.productType</em>
     * @return this
     */
    public T type(Entity type) {
        this.type = type;
        return getThis();
    }

    /**
     * Sets the product discounts.
     *
     * @param discounts the discounts
     * @return this
     */
    public T discounts(Entity... discounts) {
        this.discounts = discounts;
        return getThis();
    }

    /**
     * Builds the object.
     *
     * @param object the object
     * @param bean   a bean wrapping the object
     * @param toSave objects to save, if the object is to be saved
     */
    @Override
    protected void build(Product object, IMObjectBean bean, Set<IMObject> toSave) {
        super.build(object, bean, toSave);
        if (fixedPrice != null) {
            ProductPrice price = ProductPriceTestHelper.createFixedPrice(fixedPrice, BigDecimal.ZERO, BigDecimal.ZERO,
                                                                         BigDecimal.ZERO, (Date) null, null, true);
            object.addProductPrice(price);
        }
        if (unitPrice != null) {
            ProductPrice price = ProductPriceTestHelper.createUnitPrice(unitPrice, BigDecimal.ZERO, BigDecimal.ZERO,
                                                                        BigDecimal.ZERO, (Date) null, null);
            object.addProductPrice(price);
        }
        if (type != null) {
            bean.setTarget("type", type);
        }
        if (discounts != null) {
            for (Entity discount : discounts) {
                bean.addTarget("discounts", discount);
            }
        }
    }
}
