/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.domain.internal.object;

import org.apache.commons.collections4.iterators.AbstractUntypedIteratorDecorator;
import org.openvpms.component.business.service.archetype.helper.DefaultRelatedObjectPolicyBuilder;
import org.openvpms.component.business.service.archetype.helper.ObjectRelationshipImpl;
import org.openvpms.component.business.service.archetype.helper.RelatedIMObjects;
import org.openvpms.component.model.bean.ObjectRelationship;
import org.openvpms.component.model.bean.Policy;
import org.openvpms.component.model.bean.RelatedObjectPolicyBuilder;
import org.openvpms.component.model.bean.RelatedObjects;
import org.openvpms.component.model.object.IMObject;
import org.openvpms.component.model.object.Reference;
import org.openvpms.component.model.object.Relationship;
import org.openvpms.component.service.archetype.ArchetypeService;
import org.openvpms.domain.internal.factory.DomainService;

import java.util.Date;
import java.util.List;

/**
 * Queries related domain objects.
 *
 * @author Tim Anderson
 */
public class RelatedDomainObjects<T, R extends Relationship> implements RelatedObjects<T, R> {

    /**
     * The related objects to adapt.
     */
    private final RelatedObjects<IMObject, R> related;

    /**
     * The domain object type.
     */
    private final Class<T> type;

    /**
     * The fallback domain object type, if there is no specific domain object type for an IMObject.
     */
    private final Class<? extends T> fallbackType;

    /**
     * The domain object service.
     */
    private final DomainService domainService;

    /**
     * Constructs a {@link RelatedDomainObjects}.
     *
     * @param relationships the relationships to adapt
     * @param type          the domain object type
     * @param source        if {@code true}, return the source of the relationships, else return the target
     * @param domainService the domain object service
     * @param service       the archetype service
     */
    public RelatedDomainObjects(List<R> relationships, Class<T> type, boolean source, DomainService domainService,
                                ArchetypeService service) {
        this(relationships, type, null, source, domainService, service);
    }

    /**
     * Constructs a {@link RelatedDomainObjects}.
     *
     * @param relationships the relationships to adapt
     * @param type          the domain object type
     * @param fallbackType  the fallback domain object type, if there is no specific domain object type for an IMObject
     * @param source        if {@code true}, return the source of the relationships, else return the target
     * @param domainService the domain object service
     * @param service       the archetype service
     */
    public RelatedDomainObjects(List<R> relationships, Class<T> type, Class<? extends T> fallbackType, boolean source,
                                DomainService domainService, ArchetypeService service) {
        related = new RelatedIMObjects<>(relationships, IMObject.class, source, service);
        this.type = type;
        this.fallbackType = fallbackType;
        this.domainService = domainService;
    }

    /**
     * Selects all related objects for retrieval.
     * <p/>
     * This is the default.
     *
     * @return this
     */
    @Override
    public RelatedObjects<T, R> all() {
        related.all();
        return this;
    }

    /**
     * Selects active relationships and objects for retrieval.
     *
     * @return this
     */
    @Override
    public RelatedObjects<T, R> active() {
        related.active();
        return this;
    }

    /**
     * Selects relationships active at the specified time and returns active objects.
     *
     * @param time the time
     * @return this
     */
    @Override
    public RelatedObjects<T, R> active(Date time) {
        related.active(time);
        return this;
    }

    /**
     * Returns a policy builder to select objects.
     *
     * @return a new policy builder
     */
    @Override
    @SuppressWarnings("unchecked")
    public RelatedObjectPolicyBuilder<T, R> newPolicy() {
        return (RelatedObjectPolicyBuilder<T, R>) new DefaultRelatedObjectPolicyBuilder<>(
                (RelatedObjects<T, Relationship>) this, Relationship.class);
    }

    /**
     * Sets the policy for object retrieval.
     *
     * @param policy the policy. If {@code null}, no filtering occurs
     * @return this
     */
    @Override
    public RelatedObjects<T, R> policy(Policy<R> policy) {
        related.policy(policy);
        return this;
    }

    /**
     * Returns the first object to match the {@link #policy(Policy) policy}.
     * <ul>
     *     <li>If a {@link Policy#getComparator() comparator} is registered, this will be applied.</li>
     *     <li>If {@link Policy#getState() state}  == State.ACTIVE} the object must be active in order to be returned,
     *     otherwise an active object will be returned in preference to an inactive one.</li>
     * </ul>
     *
     * @return the first object, or {@code null} if none is found
     */
    @Override
    public T getObject() {
        IMObject object = related.getObject();
        return object != null ? domainService.create(object, type) : null;
    }

    /**
     * Returns the first object and relationship to match the {@link #policy(Policy) policy}.
     * <ul>
     *     <li>If a {@link Policy#getComparator() comparator} is registered, this will be applied.</li>
     *     <li>If {@link Policy#getState() state} == State.ACTIVE} the object must be active in order to be returned,
     *     otherwise an active object will be returned in preference to an inactive one.</li>
     * </ul>
     *
     * @return the first object and relationship, or {@code null} if none is found
     */
    @Override
    public ObjectRelationship<T, R> getObjectRelationship() {
        ObjectRelationship<IMObject, R> pair = related.getObjectRelationship();
        if (pair != null) {
            T object = domainService.create(pair.getObject(), type, fallbackType);
            return new ObjectRelationshipImpl<>(object, pair.getRelationship());
        }
        return null;
    }

    /**
     * Returns the related objects matching the {@link #policy(Policy) policy}.
     *
     * @return the related objects
     */
    @Override
    public Iterable<T> getObjects() {
        Iterable<IMObject> iterable = related.getObjects();
        return () -> new LazyObjectIterator(iterable);
    }

    /**
     * Returns the objects and their corresponding relationships matching the {@link #policy(Policy) policy}.
     *
     * @return the objects and their corresponding relationships
     */
    @Override
    public Iterable<ObjectRelationship<T, R>> getObjectRelationships() {
        Iterable<ObjectRelationship<IMObject, R>> iterable = related.getObjectRelationships();
        return () -> new LazyObjectRelationshipIterator(iterable);
    }

    /**
     * Returns the relationships matching the {@link #policy(Policy) policy}.
     *
     * @return the relationships matching the criteria
     */
    @Override
    public List<R> getRelationships() {
        return related.getRelationships();
    }

    /**
     * Returns the relationship references matching the {@link #policy(Policy) policy}.
     * <p/>
     * NOTE: this does not determine if the objects are active or inactive
     *
     * @return the related object references
     */
    @Override
    public List<Reference> getReferences() {
        return related.getReferences();
    }

    /**
     * An iterator that adapts {@link IMObject} instances to the domain object type.
     */
    private class LazyObjectIterator extends AbstractUntypedIteratorDecorator<IMObject, T> {

        /**
         * Constructs a {@link LazyObjectIterator}.
         *
         * @param iterable the iterable over the {@link IMObject} instances to adapt
         */
        public LazyObjectIterator(Iterable<IMObject> iterable) {
            super(iterable.iterator());
        }

        @Override
        public T next() {
            return domainService.create(getIterator().next(), type);
        }
    }

    /**
     * An iterator that adapts {@link ObjectRelationship} instances to relationships containing the domain object type.
     */
    private class LazyObjectRelationshipIterator
            extends AbstractUntypedIteratorDecorator<ObjectRelationship<IMObject, R>, ObjectRelationship<T, R>> {

        /**
         * Constructs a {@link LazyObjectRelationshipIterator}.
         *
         * @param iterable the iterable over the relationships to adapt
         */
        public LazyObjectRelationshipIterator(Iterable<ObjectRelationship<IMObject, R>> iterable) {
            super(iterable.iterator());
        }

        @Override
        public ObjectRelationship<T, R> next() {
            ObjectRelationship<IMObject, R> relationship = getIterator().next();
            T object = null;
            if (relationship.getObject() != null) {
                object = domainService.create(relationship.getObject(), type);
            }
            return new ObjectRelationshipImpl<>(object, relationship.getRelationship());
        }
    }

}
