/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.component.model.bean;

import org.openvpms.component.model.object.Relationship;

import java.util.Comparator;
import java.util.Date;
import java.util.function.Predicate;

/**
 * A {@link PolicyBuilder} for {@link RelatedObjects}.
 *
 * @author Tim Anderson
 */
public interface RelatedObjectPolicyBuilder<T, R extends Relationship> extends PolicyBuilder<R> {

    /**
     * Selects active relationships, and returns active objects.
     * <p/>
     * This replaces any existing predicate.
     *
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> active();

    /**
     * Selects relationships active at the specified time, and returns active objects.
     * <p/>
     * This replaces any existing predicate.
     *
     * @param time the time
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> active(Date time);

    /**
     * Selects active objects.
     *
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> activeObjects();

    /**
     * Selects inactive objects.
     *
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> inactiveObjects();

    /**
     * Selects both active and inactive objects.
     *
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> anyObject();

    /**
     * Adds a predicate for filtering relationships.
     *
     * @param predicate the predicate
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> predicate(Predicate<R> predicate);

    /**
     * Adds a predicate that is a logical AND of any existing predicate.
     *
     * @param predicate the predicate
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> and(Predicate<R> predicate);

    /**
     * Adds a predicate that is a logical OR of any existing predicate.
     *
     * @param predicate the predicate
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> or(Predicate<R> predicate);

    /**
     * Sets the comparator for ordering relationships.
     *
     * @param comparator the comparator
     * @return this
     */
    @Override
    RelatedObjectPolicyBuilder<T, R> comparator(Comparator<R> comparator);

    /**
     * Builds and registers the policy and returns the parent.
     *
     * @return the parent
     */
    RelatedObjects<T, R> add();
}
