/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.component.business.dao.hibernate.usertype;

import org.hibernate.HibernateException;
import org.hibernate.engine.spi.SharedSessionContractImplementor;
import org.hibernate.usertype.UserType;
import org.openvpms.component.business.dao.hibernate.cache.ArchetypeIdCache;
import org.openvpms.component.business.domain.archetype.ArchetypeId;

import java.io.Serializable;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;
import java.util.Objects;

/**
 * .
 *
 * @author Tim Anderson
 */
public class ArchetypeIdUserType implements UserType, Serializable {

    private ArchetypeIdCache cache = new ArchetypeIdCache();
    
    /**
     * The SQL type codes for the columns mapped by this type.
     */
    private static final int[] SQL_TYPES = {Types.VARCHAR, Types.VARCHAR};

    /**
     * Return the SQL type codes for the columns mapped by this type.
     * *
     * @return the typecodes
     */
    @Override
    public int[] sqlTypes() {
        return SQL_TYPES;
    }

    /**
     * The class returned by {@code nullSafeGet()}.
     *
     * @return Class
     */
    @Override
    public Class returnedClass() {
        return ArchetypeId.class;
    }

    /**
     * Compare two instances of the class mapped by this type for persistence "equality".
     * Equality of the persistent state.
     *
     * @param x the first argument to compare
     * @param y the second argument to compare
     * @return {@code true} if they are equal, otherwise {@code false}
     */
    @Override
    public boolean equals(Object x, Object y) {
        return Objects.equals(x, y);
    }

    /**
     * Get a hashcode for the instance, consistent with persistence "equality"
     *
     * @param x the instance
     * @return the hash code
     */
    @Override
    public int hashCode(Object x) {
        return x.hashCode();
    }

    /**
     * Retrieve an instance of the mapped class from a JDBC resultset. Implementors
     * should handle possibility of null values.
     *
     * @param rs      a JDBC result set
     * @param names   the column names
     * @param session the session
     * @param owner   the containing entity  
     * @return the archetype id. May be {@code null}
     * @throws SQLException for any SQL error
     */
    @Override
    public Object nullSafeGet(ResultSet rs, String[] names, SharedSessionContractImplementor session, Object owner)
            throws SQLException {
        String archetype = rs.getString(names[0]);
        String version = rs.getString(names[1]);
        return (archetype != null && version != null) ? cache.get(archetype, version) : null;
    }

    /**
     * Write an instance of the mapped class to a prepared statement. Implementors
     * should handle possibility of null values. A multi-column type should be written
     * to parameters starting from {@code index}.
     *
     * @param st      a JDBC prepared statement
     * @param value   the object to write
     * @param index   statement parameter index
     * @param session the session
     * @throws SQLException for any SQL error
     */
    @Override
    public void nullSafeSet(PreparedStatement st, Object value, int index, SharedSessionContractImplementor session)
            throws SQLException {
        if (value == null) {
            st.setNull(index, Types.VARCHAR);
            st.setNull(index + 1, Types.VARCHAR);
        } else {
            ArchetypeId id = (ArchetypeId) value;
            st.setString(index, id.getShortName());
            st.setString(index + 1, id.getVersion());
        }
    }

    /**
     * Return a deep copy of the persistent state, stopping at entities and at
     * collections. It is not necessary to copy immutable objects, or null
     * values, in which case it is safe to simply return the argument.
     *
     * @param value the object to be cloned, which may be null
     * @return Object a copy
     */
    @Override
    public Object deepCopy(Object value) {
        return value;
    }

    /**
     * Are objects of this type mutable?
     *
     * @return {@code false}
     */
    @Override
    public boolean isMutable() {
        return false;
    }

    /**
     * Transform the object into its cacheable representation. At the very least this
     * method should perform a deep copy if the type is mutable. That may not be enough
     * for some implementations, however; for example, associations must be cached as
     * identifier values. (optional operation)
     *
     * @param value the object to be cached
     * @return a cachable representation of the object
     */
    @Override
    public Serializable disassemble(Object value) {
        return (Serializable) value;
    }

    /**
     * Reconstruct an object from the cacheable representation. At the very least this
     * method should perform a deep copy if the type is mutable. (optional operation)
     *
     * @param cached the object to be cached
     * @param owner  the owner of the cached object
     * @return a reconstructed object from the cachable representation
     */
    @Override
    public Object assemble(Serializable cached, Object owner) {
        return cached;
    }

    /**
     * During merge, replace the existing (target) value in the entity we are merging to
     * with a new (original) value from the detached entity we are merging. For immutable
     * objects, or null values, it is safe to simply return the first parameter. For
     * mutable objects, it is safe to return a copy of the first parameter. For objects
     * with component values, it might make sense to recursively replace component values.
     *
     * @param original the value from the detached entity being merged
     * @param target   the value in the managed entity
     * @param owner the owner
     * @return the value to be merged
     */
    @Override
    public Object replace(Object original, Object target, Object owner) throws HibernateException {
        return original;
    }
}
