/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.component.business.service.archetype.helper;

import org.openvpms.component.model.bean.AbstractPolicyBuilder;
import org.openvpms.component.model.bean.Policy;
import org.openvpms.component.model.bean.RelatedObjectPolicyBuilder;
import org.openvpms.component.model.bean.RelatedObjects;
import org.openvpms.component.model.object.Relationship;

import java.util.Comparator;
import java.util.Date;
import java.util.function.Predicate;

/**
 * Default implementation of {@link RelatedObjectPolicyBuilder}.
 *
 * @author Tim Anderson
 */
public class DefaultRelatedObjectPolicyBuilder<T, R extends Relationship> extends AbstractPolicyBuilder<R>
        implements RelatedObjectPolicyBuilder<T, R> {

    /**
     * The parent.
     */
    private final RelatedObjects<T, R> parent;

    /**
     * Constructs an {@link DefaultRelatedObjectPolicyBuilder}.
     *
     * @param parent the parent
     * @param type   the relationship type
     */
    public DefaultRelatedObjectPolicyBuilder(RelatedObjects<T, R> parent, Class<R> type) {
        super(type);
        this.parent = parent;
    }

    /**
     * Selects active relationships, and returns active objects.
     * <p/>
     * This replaces any existing predicate.
     *
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> active() {
        return (RelatedObjectPolicyBuilder<T, R>) super.active();
    }

    /**
     * Selects relationships active at the specified time, and returns active objects.
     * <p/>
     * This replaces any existing predicate.
     *
     * @param time the time
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> active(Date time) {
        return (RelatedObjectPolicyBuilder<T, R>) super.active(time);
    }

    /**
     * Selects active objects.
     *
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> activeObjects() {
        return (RelatedObjectPolicyBuilder<T, R>) super.activeObjects();
    }

    /**
     * Selects inactive objects.
     *
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> inactiveObjects() {
        return (RelatedObjectPolicyBuilder<T, R>) super.inactiveObjects();
    }

    /**
     * Selects both active and inactive objects.
     *
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> anyObject() {
        return (RelatedObjectPolicyBuilder<T, R>) super.anyObject();
    }

    /**
     * Sets the predicate for filtering relationships.
     *
     * @param predicate the predicate
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> predicate(Predicate<R> predicate) {
        return (RelatedObjectPolicyBuilder<T, R>) super.predicate(predicate);
    }

    /**
     * Adds a predicate that is a logical AND of any existing predicate.
     *
     * @param predicate the predicate
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> and(Predicate<R> predicate) {
        return (RelatedObjectPolicyBuilder<T, R>) super.and(predicate);
    }

    /**
     * Adds a predicate that is a logical OR of any existing predicate.
     *
     * @param predicate the predicate
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> or(Predicate<R> predicate) {
        return (RelatedObjectPolicyBuilder<T, R>) super.or(predicate);
    }

    /**
     * Sets the comparator for ordering relationships.
     *
     * @param comparator the comparator
     * @return this
     */
    @Override
    public RelatedObjectPolicyBuilder<T, R> comparator(Comparator<R> comparator) {
        return (RelatedObjectPolicyBuilder<T, R>) super.comparator(comparator);
    }

    /**
     * Builds and registers the policy and returns the parent.
     *
     * @return the parent
     */
    @Override
    public RelatedObjects<T, R> add() {
        Policy<R> policy = build();
        return parent.policy(policy);
    }
}
