/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.laboratory.internal.report;

import org.apache.commons.lang3.StringUtils;
import org.openvpms.archetype.rules.patient.InvestigationArchetypes;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.laboratory.exception.LaboratoryException;
import org.openvpms.laboratory.internal.i18n.LaboratoryMessages;
import org.openvpms.laboratory.report.Result;
import org.openvpms.laboratory.report.ResultBuilder;
import org.openvpms.laboratory.report.ResultsBuilder;

import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.util.List;

/**
 * Default implementation of {@link ResultBuilder}.
 *
 * @author Tim Anderson
 */
class ResultBuilderImpl extends Builder<Result> implements ResultBuilder {

    /**
     * The objects to save.
     */
    private final List<Act> toSave;

    /**
     * The act to update.
     */
    private IMObjectBean bean;

    /**
     * The result identifier.
     */
    private String resultId;

    /**
     * The result status.
     */
    private Result.Status status;

    /**
     * The analyte code.
     */
    private String analyteCode;

    /**
     * The analyte name.
     */
    private String analyteName;

    /**
     * The numeric result for the analyte.
     */
    private BigDecimal value;

    /**
     * The text result for the analyte.
     */
    private String result;

    /**
     * The units of measure for the result.
     */
    private String units;

    /**
     * The qualifier.
     */
    private String qualifier;

    /**
     * The low reference range value for the analyte.
     */
    private BigDecimal lowRange;

    /**
     * The high reference range value for the analyte.
     */
    private BigDecimal highRange;

    /**
     * The extreme low reference range value for the analyte.
     */
    private BigDecimal extremeLowRange;

    /**
     * The extreme high reference range value for the analyte.
     */
    private BigDecimal extremeHighRange;

    /**
     * Determines if the result is out of range. May be {@code null}
     */
    private Boolean outOfRange;

    /**
     * The reference range.
     */
    private String referenceRange;

    /**
     * The notes.
     */
    private String notes;

    /**
     * Constructs a {@link ResultBuilderImpl}.
     *
     * @param parent  the parent builder
     * @param id      the result identifier
     * @param results the parent results act, if it is being updated, otherwise {@code null}
     * @param state   the builder state
     * @param toSave  the objects to save
     */
    ResultBuilderImpl(ResultsBuilderImpl parent, String id, Act results, State state, List<Act> toSave) {
        super(parent);
        this.toSave = toSave;
        bean = (results != null) ? state.getResultItem(results, id) : null;
        if (bean != null) {
            extract(bean);
        } else {
            reset();
            resultId = id;
        }
    }

    /**
     * Sets the result status.
     *
     * @param status the result status
     * @return this
     */
    @Override
    public ResultBuilder status(Result.Status status) {
        this.status = status;
        return this;
    }

    /**
     * Sets the identifier for the analyte.
     *
     * @param code the analyte identifier. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder analyteCode(String code) {
        this.analyteCode = StringUtils.trimToNull(code);
        return this;
    }

    /**
     * Sets the name of the analyte for which the result is being reported.
     *
     * @param name the analyte name
     * @return this
     */
    @Override
    public ResultBuilder analyteName(String name) {
        this.analyteName = StringUtils.trimToNull(name);
        return this;
    }

    /**
     * Sets the numerical result for the analyte.
     *
     * @param value the numerical result for the analyte. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder value(BigDecimal value) {
        this.value = value;
        return this;
    }

    /**
     * Sets the result for the analyte, as text.
     *
     * @param result the result text. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder result(String result) {
        this.result = StringUtils.trimToNull(result);
        return this;
    }

    /**
     * Sets the unit of measure for the analyte result.
     *
     * @param units the unit of measure. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder units(String units) {
        this.units = StringUtils.trimToNull(units);
        return this;
    }

    /**
     * Sets the qualifier.
     *
     * @param qualifier the qualifier. May be {@code null}
     */
    @Override
    public ResultBuilder qualifier(String qualifier) {
        this.qualifier = StringUtils.trimToNull(qualifier);
        return this;
    }

    /**
     * Sets the low reference range value for the analyte.
     *
     * @param lowRange the low reference range value. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder lowRange(BigDecimal lowRange) {
        this.lowRange = lowRange;
        return this;
    }

    /**
     * Sets the high reference range value for the analyte.
     *
     * @param highRange the high reference range value. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder highRange(BigDecimal highRange) {
        this.highRange = highRange;
        return this;
    }

    /**
     * Sets the extreme low reference range value for the analyte.
     *
     * @param extremeLowRange the extreme low reference range value. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder extremeLowRange(BigDecimal extremeLowRange) {
        this.extremeLowRange = extremeLowRange;
        return this;
    }

    /**
     * Sets the extreme high reference range value for the analyte.
     *
     * @param extremeHighRange the extreme high reference range value. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder extremeHighRange(BigDecimal extremeHighRange) {
        this.extremeHighRange = extremeHighRange;
        return this;
    }

    /**
     * Determines if the result is out of range.
     *
     * @param outOfRange if {@code true}, the result is out of range
     */
    @Override
    public ResultBuilder outOfRange(boolean outOfRange) {
        this.outOfRange = outOfRange;
        return this;
    }

    /**
     * Sets the reference range.
     *
     * @param range the reference range. May be {@code null}
     */
    @Override
    public ResultBuilder referenceRange(String range) {
        this.referenceRange = StringUtils.trimToNull(range);
        return this;
    }

    /**
     * Sets a short narrative of the result.
     *
     * @param notes the result narrative. May be {@code null}
     * @return this
     */
    @Override
    public ResultBuilder notes(String notes) {
        this.notes = StringUtils.trimToNull(notes);
        return this;
    }

    /**
     * Builds the result and adds it to the set of results.
     * <p/>
     * The builder is reset.
     *
     * @return the result
     */
    @Override
    public Result build() {
        return new ResultImpl(buildResult());
    }

    /**
     * Builds the result and adds it to the set of results.
     *
     * @return the parent builder
     */
    @Override
    public ResultsBuilder add() {
        buildResult();
        return (ResultsBuilder) getParent();
    }

    /**
     * Populates a result.
     *
     * @param bean the result to populate
     */
    protected void populate(IMObjectBean bean) {
        bean.setValue("resultId", resultId);
        if (status != null) {
            bean.setValue("status", status.toString());
        }
        bean.setValue("analyteCode", analyteCode);
        bean.setValue("name", analyteName);
        bean.setValue("value", value);
        bean.setValue("result", getResult());
        bean.setValue("units", units);
        bean.setValue("qualifier", qualifier);
        bean.setValue("lowRange", lowRange);
        bean.setValue("highRange", highRange);
        bean.setValue("extremeLowRange", extremeLowRange);
        bean.setValue("extremeHighRange", extremeHighRange);
        bean.setValue("outOfRange", getOutOfRange());
        bean.setValue("referenceRange", getReferenceRange());
        bean.setValue("notes", notes);
    }

    /**
     * Use the values from the specified result.
     *
     * @param bean the result bean
     */
    private void extract(IMObjectBean bean) {
        resultId = bean.getString("resultId");
        status = ResultImpl.getStatus(bean.getString("status"));
        analyteCode = bean.getString("analyteCode");
        analyteName = bean.getString("name");
        value = bean.getBigDecimal("value");
        result = bean.getString("result");
        units = bean.getString("units");
        qualifier = bean.getString("qualifier");
        lowRange = bean.getBigDecimal("lowRange");
        highRange = bean.getBigDecimal("highRange");
        extremeLowRange = bean.getBigDecimal("extremeLowRange");
        extremeHighRange = bean.getBigDecimal("extremeHighRange");
        outOfRange = (Boolean) bean.getValue("outOfRange");
        referenceRange = bean.getString("referenceRange");
        notes = bean.getString("notes");
    }

    /**
     * Resets the builder.
     */
    private void reset() {
        resultId = null;
        status = null;
        analyteCode = null;
        analyteName = null;
        value = null;
        result = null;
        units = null;
        qualifier = null;
        lowRange = null;
        highRange = null;
        extremeLowRange = null;
        extremeHighRange = null;
        outOfRange = null;
        referenceRange = null;
        notes = null;
    }

    /**
     * Returns the result, deriving it from the value if none is specified.
     *
     * @return the result. May be {@code null}
     */
    private String getResult() {
        if (result == null && value != null) {
            return new DecimalFormat("#,###.#####").format(value);
        }
        return result;
    }

    /**
     * Determines if the result is out of range, if no flag was specified, but the value and either
     * the extreme low/high or low/high ranges are.
     *
     * @return {@code true} if the result is out of range, {@code false} if it is in range, or {@code null} if it cannot
     * be determined
     */
    private Boolean getOutOfRange() {
        if (outOfRange == null && value != null) {
            if (extremeLowRange != null && extremeHighRange != null) {
                return (value.compareTo(extremeLowRange) < 0 || value.compareTo(extremeHighRange) > 0);
            } else if (lowRange != null && highRange != null) {
                return (value.compareTo(lowRange) < 0 || value.compareTo(highRange) > 0);
            }
        }
        return outOfRange;
    }

    /**
     * Returns the reference range, deriving it from the extreme low/high range or low/high range and units if none is
     * specified.
     *
     * @return the reference range. May be {@code null}
     */
    private String getReferenceRange() {
        String result = referenceRange;
        if (result == null) {
            if (extremeLowRange != null && extremeHighRange != null) {
                result = getReferenceRange(extremeLowRange, extremeHighRange);
            } else if (lowRange != null && highRange != null) {
                result = getReferenceRange(lowRange, highRange);
            }
        }
        return result;
    }

    /**
     * Formats a reference range.
     *
     * @param low  the low value
     * @param high the high value
     * @return the formatted range
     */
    private String getReferenceRange(BigDecimal low, BigDecimal high) {
        DecimalFormat format = new DecimalFormat("#,###.#####");
        StringBuilder builder = new StringBuilder();
        builder.append(format.format(low))
                .append(" - ")
                .append(format.format(high));
        if (units != null) {
            builder.append(' ').append(units);
        }
        return builder.toString();
    }

    /**
     * Builds the result.
     *
     * @return the result bean
     */
    private IMObjectBean buildResult() {
        if (resultId == null) {
            throw new LaboratoryException(LaboratoryMessages.noResultId());
        }
        Act act;
        if (bean == null) {
            act = create(InvestigationArchetypes.RESULT, Act.class);
            bean = getBean(act);
        } else {
            act = (Act) bean.getObject();
        }
        populate(bean);
        toSave.add(act);
        IMObjectBean result = bean;
        reset();
        return result;
    }

}
