/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.laboratory.internal.report;

import org.openvpms.archetype.rules.workflow.WorkflowStatus;
import org.openvpms.component.business.domain.im.act.BeanActDecorator;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.service.archetype.ArchetypeService;
import org.openvpms.laboratory.report.Result;

import java.math.BigDecimal;

/**
 * Default implementation of {@link Result}.
 *
 * @author Tim Anderson
 */
public class ResultImpl implements Result {

    /**
     * The underlying act.
     */
    private BeanActDecorator act;

    /**
     * Constructs a {@link ResultImpl}.
     *
     * @param act     the peer to delegate to
     * @param service the archetype service
     */
    public ResultImpl(Act act, ArchetypeService service) {
        this.act = new BeanActDecorator(act, service);
    }

    /**
     * Creates a {@link ResultImpl}.
     *
     * @param bean the bean wrapping the act
     */
    ResultImpl(IMObjectBean bean) {
        this.act = new BeanActDecorator(bean);
    }

    /**
     * Returns the result identifier.
     *
     * @return the result identifier
     */
    @Override
    public String getResultId() {
        return act.getBean().getString("resultId");
    }

    /**
     * Returns the result status.
     *
     * @return the result status
     */
    @Override
    public Status getStatus() {
        return getStatus(act.getStatus());
    }

    /**
     * Returns an identifier for the analyte.
     *
     * @return the analyte identifier. May be {@code null}
     */
    @Override
    public String getAnalyteCode() {
        return act.getBean().getString("analyteCode");
    }

    /**
     * Returns the name of the analyte for which the result is being reported.
     *
     * @return the analyte name
     */
    @Override
    public String getAnalyteName() {
        return act.getBean().getString("name");
    }

    /**
     * Returns the numerical result for the analyte.
     *
     * @return the numerical result for the analyte. May be {@code null}
     */
    @Override
    public BigDecimal getValue() {
        return act.getBean().getBigDecimal("value");
    }

    /**
     * Returns the result for the analyte, as text.
     *
     * @return the result text. May be {@code null}
     */
    @Override
    public String getResult() {
        return act.getBean().getString("result");
    }

    /**
     * Returns the unit of measure for the analyte result.
     *
     * @return the unit of measure. May be {@code null}
     */
    @Override
    public String getUnits() {
        return act.getBean().getString("units");
    }

    /**
     * Returns the qualifier.
     *
     * @return the qualifier. May be {@code null}
     */
    @Override
    public String getQualifier() {
        return act.getBean().getString("qualifier");
    }

    /**
     * Returns the low reference range value for the analyte.
     *
     * @return the low reference range value. May be {@code null}
     */
    @Override
    public BigDecimal getLowRange() {
        return act.getBean().getBigDecimal("lowRange");
    }

    /**
     * Returns the high reference range value for the analyte.
     *
     * @return the high reference range value. May be {@code null}
     */
    @Override
    public BigDecimal getHighRange() {
        return act.getBean().getBigDecimal("highRange");
    }

    /**
     * Returns the extreme low reference range value for the analyte.
     *
     * @return the extreme low reference range value. May be {@code null}
     */
    @Override
    public BigDecimal getExtremeLowRange() {
        return act.getBean().getBigDecimal("extremeLowRange");
    }

    /**
     * Returns the extreme high reference range value for the analyte.
     *
     * @return the extreme high reference range value. May be {@code null}
     */
    @Override
    public BigDecimal getExtremeHighRange() {
        return act.getBean().getBigDecimal("extremeHighRange");
    }

    /**
     * Determines if the result is out of range.
     *
     * @return {@code true} if the result is out of range, otherwise {@code false}
     */
    @Override
    public boolean getOutOfRange() {
        return act.getBean().getBoolean("outOfRange");
    }

    /**
     * Sets the reference range.
     *
     * @return the reference range. May be {@code null}
     */
    @Override
    public String getReferenceRange() {
        return act.getBean().getString("referenceRange");
    }

    /**
     * Returns a short narrative of the result.
     *
     * @return the result narrative. May be {@code null}
     */
    @Override
    public String getNotes() {
        return act.getBean().getString("notes");
    }

    /**
     * Converts a status string to an enum.
     *
     * @param status the status string
     * @return the corresponding status
     */
    static Status getStatus(String status) {
        if (WorkflowStatus.IN_PROGRESS.equals(status)) {
            return Status.IN_PROGRESS;
        } else if (WorkflowStatus.COMPLETED.equals(status)) {
            return Status.COMPLETED;
        } else if (WorkflowStatus.CANCELLED.equals(status)) {
            return Status.CANCELLED;
        }
        return Status.PENDING;
    }
}
