/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.sms.internal.message;

import org.openvpms.archetype.rules.util.DateRules;
import org.openvpms.component.business.domain.im.act.ActIdentity;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.party.Party;
import org.openvpms.component.service.archetype.ArchetypeService;
import org.openvpms.domain.internal.factory.DomainService;
import org.openvpms.domain.practice.Location;
import org.openvpms.sms.internal.SMSArchetypes;
import org.openvpms.sms.message.Message;
import org.openvpms.sms.message.Reply;
import org.openvpms.sms.message.ReplyBuilder;
import org.openvpms.sms.message.StateBuilder;

import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * Default implementation of {@link Message}.
 *
 * @author Tim Anderson
 */
public class MessageImpl implements Message {

    /**
     * The queue status.
     */
    public enum QueueStatus {
        PENDING,
        QUEUED,
        COMPLETED,
    }

    /**
     * The underlying message.
     */
    private final Act message;

    /**
     * The bean wrapping the message.
     */
    private final IMObjectBean bean;

    /**
     * The archetype service.
     */
    private final ArchetypeService service;

    /**
     * The domain object service.
     */
    private final DomainService domainService;


    /**
     * Constructs a {@link MessageImpl}.
     *
     * @param message       the message
     * @param service       the archetype service
     * @param domainService the domain object service
     */
    public MessageImpl(Act message, ArchetypeService service, DomainService domainService) {
        this.message = message;
        this.bean = service.getBean(message);
        this.service = service;
        this.domainService = domainService;
    }

    /**
     * Returns the OpenVPMS identifier for this message.
     *
     * @return the identifier
     */
    @Override
    public long getId() {
        return message.getId();
    }

    /**
     * Returns the message identifier, issued by the SMS provider.
     *
     * @return the message identifier, or {@code null} if none has been issued
     */
    @Override
    public String getProviderId() {
        ActIdentity identity = getIdentity();
        return identity != null ? identity.getIdentity() : null;
    }

    /**
     * Returns the timestamp when the message was created.
     *
     * @return the timestamp
     */
    @Override
    public OffsetDateTime getCreated() {
        return DateRules.toOffsetDateTime(bean.getDate("createdTime"));
    }

    /**
     * Returns the timestamp when the message was updated.
     *
     * @return the timestamp
     */
    @Override
    public OffsetDateTime getUpdated() {
        return DateRules.toOffsetDateTime(bean.getDate("updatedTime"));
    }

    /**
     * Returns the timestamp when the message expires.
     * <p/>
     * It will no longer be sent after this.
     *
     * @return the timestamp, or {@code null} if the message never expires
     */
    @Override
    public OffsetDateTime getExpiry() {
        Date date = getExpiryDate();
        return (date != null) ? DateRules.toOffsetDateTime(date) : null;
    }

    /**
     * Determines if the message has expired.
     * <p/>
     * This only applies to messages that have {@code PENDING status}.
     *
     * @return {@code true} if the message has expired
     */
    @Override
    public boolean isExpired() {
        boolean result = false;
        Status status = getStatus();
        Date date = getExpiryDate();
        if (date != null && status == Status.PENDING) {
            result = date.before(new Date());
        }
        return result;
    }

    /**
     * Returns the timestamp when the message was sent.
     *
     * @return the timestamp when the message was sent, or {@code null} if it hasn't been sent
     */
    @Override
    public OffsetDateTime getSent() {
        Date date = bean.getDate("endTime");
        return (date != null) ? DateRules.toOffsetDateTime(date) : null;
    }

    /**
     * Returns the party that this message is for.
     *
     * @return the party, or {@code null} if the party is not known
     */
    @Override
    public Party getRecipient() {
        return bean.getTarget("recipient", Party.class);
    }

    /**
     * Returns the phone number.
     *
     * @return the phone number
     */
    @Override
    public String getPhone() {
        return bean.isA(SMSArchetypes.MESSAGE) ? bean.getString("phone") : bean.getString("address");
    }

    /**
     * Returns the message text.
     *
     * @return the message text
     */
    @Override
    public String getMessage() {
        return bean.getString("message");
    }

    /**
     * Returns the message status.
     *
     * @return the message status
     */
    @Override
    public Status getStatus() {
        return Status.valueOf(message.getStatus());
    }

    /**
     * Sets the message status.
     * <p/>
     * This clears any status message.
     *
     * @param status the message status
     */
    @Override
    public void setStatus(Status status) {
        setStatus(status, null);
    }

    /**
     * Returns the status message.
     *
     * @return the status message. May be {@code null}
     */
    @Override
    public String getStatusMessage() {
        return bean.getString("statusMessage");
    }

    /**
     * Sets the message status, along with any status message from the SMS provider.
     *
     * @param status        the status
     * @param statusMessage the status message. May be {@code null}
     */
    @Override
    public void setStatus(Status status, String statusMessage) {
        state().status(status, statusMessage).build();
    }

    /**
     * Determines if the message can transition to the specified status.
     *
     * @param status the status
     * @return {@code true} if the message can transition to the specified status, otherwise {@code false}
     */
    @Override
    public boolean canTransition(Status status) {
        Status current = getStatus();
        return current.canTransition(status);
    }

    /**
     * Determines if the message has a reply with the specified identifier.
     *
     * @param providerId the provider identifier
     * @return {@code true} if there is a reply with the specified identifier
     */
    @Override
    public boolean hasReply(String providerId) {
        for (Reply reply : getReplies()) {
            if (providerId.equals(reply.getProviderId())) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns any replies to the message.
     *
     * @return the replies
     */
    @Override
    public List<Reply> getReplies() {
        List<Reply> result = new ArrayList<>();
        for (Act act : bean.getTargets("replies", Act.class)) {
            result.add(new ReplyImpl(act, service));
        }
        return result;
    }

    /**
     * Returns the practice location that this message originated from.
     *
     * @return the practice location
     */
    @Override
    public Location getLocation() {
        Party location = bean.getTarget("location", Party.class);
        return domainService.create(location, Location.class);
    }

    /**
     * Returns a builder to update the state of the message.
     *
     * @return a builder to update the message
     */
    @Override
    public StateBuilder state() {
        return new StateBuilderImpl(this, bean, service);
    }

    /**
     * Returns a builder to add a reply to the message.
     *
     * @return a builder to add a reply
     */
    @Override
    public ReplyBuilder reply() {
        return state().reply();
    }

    /**
     * Queues the message for send to the provider.
     */
    public void queue() {
        if (getStatus() != Message.Status.PENDING) {
            throw new IllegalStateException("Cannot queue message with status=" + message.getStatus());
        }
        if (getQueueStatus() == QueueStatus.PENDING) {
            message.setStatus2(QueueStatus.QUEUED.toString());
            setUpdated();
            bean.save();
        }
    }

    /**
     * Returns the queue status.
     *
     * @return the queue status
     */
    public QueueStatus getQueueStatus() {
        return QueueStatus.valueOf(message.getStatus2());
    }

    /**
     * Returns the <em>expiryTime</em> node as a Date.
     *
     * @return the expiry time, or {@code null} if the message doesn't expire
     */
    protected Date getExpiryDate() {
        return bean.getDate("expiryTime");
    }

    /**
     * Returns the message identity, as specified by the SMS provider.
     *
     * @return the message identity, or {@code null} if none is registered
     */
    protected ActIdentity getIdentity() {
        return bean.getObject("providerId", ActIdentity.class);
    }

    /**
     * Sets the updated node.
     */
    void setUpdated() {
        bean.setValue("startTime", new Date());
    }

}