/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.sms.internal.message;

import org.openvpms.archetype.rules.util.DateRules;
import org.openvpms.component.business.domain.im.act.ActIdentity;
import org.openvpms.component.business.domain.im.act.BeanActDecorator;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.service.archetype.ArchetypeService;
import org.openvpms.sms.message.Reply;

import java.time.OffsetDateTime;

/**
 * Default implementation of {@link Reply}.
 *
 * @author Tim Anderson
 */
public class ReplyImpl implements Reply {

    /**
     * The underlying act.
     */
    private final BeanActDecorator act;

    /**
     * Constructs a {@link ReplyImpl}.
     *
     * @param reply   the act
     * @param service the archetype service
     */
    public ReplyImpl(Act reply, ArchetypeService service) {
        act = new BeanActDecorator(reply, service);
    }

    /**
     * Constructs a {@link ReplyImpl}.
     *
     * @param bean a bean wrapping the act
     */
    public ReplyImpl(IMObjectBean bean) {
        act = new BeanActDecorator(bean);
    }

    /**
     * Returns the OpenVPMS identifier for this reply.
     *
     * @return the identifier
     */
    @Override
    public long getId() {
        return act.getId();
    }

    /**
     * Returns the reply identifier, issued by the SMS provider.
     *
     * @return the reply identifier, or {@code null} if none has been issued
     */
    @Override
    public String getProviderId() {
        ActIdentity identity = act.getBean().getObject("providerId", ActIdentity.class);
        return identity != null ? identity.getIdentity() : null;
    }

    /**
     * Returns the time when the reply was received.
     *
     * @return the reply receipt time
     */
    @Override
    public OffsetDateTime getReceived() {
        return DateRules.toOffsetDateTime(act.getBean().getDate("startTime"));
    }

    /**
     * Returns the phone number.
     *
     * @return the phone number
     */
    @Override
    public String getPhone() {
        return act.getBean().getString("phone");
    }

    /**
     * Returns the message text.
     *
     * @return the message text
     */
    @Override
    public String getMessage() {
        return act.getBean().getString("message");
    }

}
