/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.sms.internal.mail.message;

import org.junit.Before;
import org.junit.Test;
import org.openvpms.archetype.rules.util.DateRules;
import org.openvpms.archetype.test.ArchetypeServiceTest;
import org.openvpms.archetype.test.TestHelper;
import org.openvpms.component.model.act.Act;
import org.openvpms.component.model.bean.IMObjectBean;
import org.openvpms.component.model.party.Party;
import org.openvpms.domain.internal.factory.DomainService;
import org.openvpms.domain.internal.factory.DomainServiceImpl;
import org.openvpms.sms.internal.SMSArchetypes;
import org.openvpms.sms.internal.message.MessageBuilderImpl;
import org.openvpms.sms.internal.message.MessageImpl;
import org.openvpms.sms.message.Message;
import org.openvpms.sms.message.Message.Status;

import java.time.OffsetDateTime;
import java.util.Date;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

/**
 * Tests the {@link MessageBuilderImpl} class.
 *
 * @author Tim Anderson
 */
public class MessageBuilderImplTestCase extends ArchetypeServiceTest {

    /**
     * The builder.
     */
    private MessageBuilderImpl builder;

    /**
     * Sets up the test case.
     */
    @Before
    public void setUp() {
        DomainService domainService = new DomainServiceImpl(getArchetypeService());
        builder = new MessageBuilderImpl(getArchetypeService(), domainService);
    }

    /**
     * Tests the building a message.
     */
    @Test
    public void testBuild() {
        Party recipient = TestHelper.createSupplierVet();
        Party location = TestHelper.createLocation();
        Party customer = TestHelper.createCustomer();
        Party patient = TestHelper.createPatient();
        String reason = TestHelper.getLookup("lookup.customerCommunicationReason", "PATIENT_UPDATE").getCode();
        Date expiryDate = DateRules.getTomorrow();
        OffsetDateTime expiryDateTime = DateRules.toOffsetDateTime(expiryDate);
        Message message = builder.recipient(recipient)
                .phone("12345678")
                .message("a message")
                .location(location)
                .reason(reason)
                .subject("a subject")
                .patient(patient)
                .customer(customer)
                .expiry(expiryDateTime)
                .note("a note")
                .build();

        assertEquals(recipient, message.getRecipient());
        assertEquals("12345678", message.getPhone());
        assertEquals("a message", message.getMessage());
        assertEquals(Status.PENDING, message.getStatus());
        assertEquals(expiryDateTime, message.getExpiry());

        ((MessageImpl) message).queue();  // required to save it.

        // check the underlying act
        Act act = getAct(message);
        IMObjectBean bean = getBean(act);
        assertEquals(recipient, bean.getTarget("recipient"));
        assertEquals("12345678", bean.getString("phone"));
        assertEquals("a message", bean.getString("message"));
        assertEquals(Status.PENDING.toString(), act.getStatus());
        assertEquals(reason, act.getReason());
        assertEquals(expiryDate, bean.getDate("expiryTime"));
        assertEquals(MessageImpl.QueueStatus.QUEUED.toString(), act.getStatus2());
        assertEquals(customer, bean.getTarget("customer"));
        assertEquals(patient, bean.getTarget("patient"));
        assertEquals("a note", bean.getString("note"));
    }

    /**
     * Verifies that when the recipient is a customer, the customer node is automatically populated.
     */
    @Test
    public void testCustomerIsSetWhenRecipientIsACustomer() {
        Party customer = TestHelper.createCustomer();
        Message message = builder.recipient(customer)
                .phone("12345678")
                .message("a message")
                .build();

        ((MessageImpl) message).queue();  // required to save it.

        // check the underlying act
        Act act = getAct(message);
        IMObjectBean bean = getBean(act);
        assertEquals(customer, bean.getTarget("recipient"));
        assertEquals(customer, bean.getTarget("customer"));
    }

    /**
     * Verifies a message can be built with no recipient.
     */
    @Test
    public void testNoRecipient() {
        Party customer = TestHelper.createCustomer();
        Message message = builder.customer(customer)
                .phone("12345678")
                .message("a message")
                .build();

        ((MessageImpl) message).queue();  // required to save it.

        // check the underlying act
        Act act = getAct(message);
        IMObjectBean bean = getBean(act);
        assertNull(bean.getTarget("recipient"));
        assertEquals(customer, bean.getTarget("customer"));
    }

    /**
     * Returns the <em>act.smsMessage/em>> associated with a {@link Message}.
     *
     * @param message the message. Must be saved
     * @return the corresponding
     */
    private Act getAct(Message message) {
        Act act = getArchetypeService().get(SMSArchetypes.MESSAGE, message.getId(), Act.class);
        assertNotNull(act);
        return act;
    }
}
