/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.tool.toolbox.config;

import org.apache.commons.lang3.StringUtils;

import java.io.FileInputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collections;
import java.util.Enumeration;
import java.util.LinkedHashSet;
import java.util.Properties;
import java.util.Set;

/**
 * The config properties.
 *
 * @author Tim Anderson
 */
public class ConfigProperties {

    /**
     * JDBC driver class name property.
     */
    protected static final String DB_DRIVER = "db.driver";

    /**
     * JDBC url property.
     */
    protected static final String DB_URL = "db.url";

    /**
     * JDBC username property.
     */
    protected static final String DB_USERNAME = "db.username";

    /**
     * JDBC password property.
     */
    protected static final String DB_PASSWORD = "db.password";

    /**
     * Reporting JDBC URL property.
     */
    protected static final String REPORTING_DB_URL = "reporting.db.url";

    /**
     * Reporting JDBC username property.
     */
    protected static final String REPORTING_DB_USERNAME = "reporting.db.username";

    /**
     * Reporting JDBC password property.
     */
    protected static final String REPORTING_DB_PASSWORD = "reporting.db.password";

    /**
     * Encryption key property.
     */
    protected static final String OPENVPMS_KEY = "openvpms.key";

    /**
     * The property file path.
     */
    private final Path path;

    /**
     * The properties.
     */
    private final Properties properties;

    /**
     * The properties.
     */
    private static final String[] PROPERTIES = new String[]{DB_DRIVER, DB_URL, DB_USERNAME, DB_PASSWORD,
                                                            REPORTING_DB_URL, REPORTING_DB_USERNAME,
                                                            REPORTING_DB_PASSWORD, OPENVPMS_KEY};

    /**
     * Creates a {@link ConfigProperties}.
     *
     * @param path the configuration file path
     * @throws IOException for any I/O error
     */
    public ConfigProperties(Path path) throws IOException {
        this.path = path;
        Properties properties = new OrderedProperties();
        if (Files.exists(path)) {
            properties.load(new FileInputStream(path.toFile()));
        }
        this.properties = properties;
    }

    /**
     * Returns the JDBC driver class name.
     *
     * @return the JDBC driver class name. May be {@code null}
     */
    public String getDriver() {
        return get(DB_DRIVER);
    }

    /**
     * Returns the database URL.
     *
     * @return the database URL. May be {@code null}
     */
    public String getUrl() {
        return get(DB_URL);
    }

    /**
     * Returns the database user name.
     *
     * @return the database user name. May be {@code null}
     */
    public String getUsername() {
        return get(DB_USERNAME);
    }

    /**
     * Returns the database password.
     *
     * @return the database password. May be {@code null}
     */
    public String getPassword() {
        return get(DB_PASSWORD);
    }

    /**
     * Returns the reporting database URL.
     *
     * @return the reporting database URL. May be {@code null}
     */
    public String getReportingUrl() {
        return get(REPORTING_DB_URL);
    }

    /**
     * Returns the reporting user name.
     *
     * @return the reporting user name. May be {@code null}
     */
    public String getReportingUsername() {
        return get(REPORTING_DB_USERNAME);
    }

    /**
     * Returns the reporting password.
     *
     * @return the reporting password. May be {@code null}
     */
    public String getReportingPassword() {
        return get(REPORTING_DB_PASSWORD);
    }

    /**
     * Returns the encryption key.
     *
     * @return the encryption key. May be {@code null}
     */
    public String getKey() {
        return get(OPENVPMS_KEY);
    }

    /**
     * Determines if the configuration is complete.
     *
     * @return {@code true} if the configuration is complete
     */
    public boolean isComplete() {
        for (String name : PROPERTIES) {
            if (get(name) == null) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns the properties path.
     *
     * @return the properties path
     */
    public Path getPath() {
        return path;
    }

    /**
     * Returns the properties.
     *
     * @return the properties
     */
    public Properties getProperties() {
        return properties;
    }

    /**
     * Returns a property.
     *
     * @param name the property name
     * @return the property value. May be {@code null}
     */
    private String get(String name) {
        return StringUtils.trimToNull(properties.getProperty(name));
    }

    /**
     * Ensures that property ordering is preserved when writing properties.
     */
    private static class OrderedProperties extends Properties {

        /**
         * The keys.
         */
        private final Set<Object> keys = new LinkedHashSet<>();

        @Override
        public Enumeration<Object> keys() {
            return Collections.enumeration(keys);
        }

        @Override
        public Object put(Object key, Object value) {
            keys.add(key);
            return super.put(key, value);
        }
    }
}
