/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.tool.toolbox.db;

import org.openvpms.tool.db.DBTool;
import picocli.CommandLine;
import picocli.CommandLine.Command;

import java.io.Console;

/**
 * Command to update the database to the latest version.
 *
 * @author Tim Anderson
 */
@Command(name = "--update", header = "Updates the database to the latest version")
public class UpdateDBCommand extends AbstractDBCommand {

    /**
     * Flag to override interactive prompt.
     */
    @CommandLine.Option(names = "--database-is-backed-up", hidden = true)
    boolean backedup;

    /**
     * Runs the command.
     *
     * @return {@code 0} indicates success, non-zero failure
     * @throws Exception for any error
     */
    @Override
    protected int run() throws Exception {
        int result = 1;
        DBTool tool = getTool();
        String key = getProperties().getKey();
        if (tool.needsUpdate()) {
            if (backedup()) {
                System.setProperty("openvpms.key", key);
                tool.update();
                result = 0;
            }
        } else {
            System.out.println("Database is up to date");
            result = 0;
        }
        return result;
    }

    /**
     * Determines if the database is backed up.
     *
     * @return {@code true} if the database is backed up, otherwise {@code false}
     */
    private boolean backedup() {
        boolean result = false;
        if (!backedup) {
            Console console = System.console();
            if (console == null) {
                System.err.println("This command must be executed in a console");
            } else {
                while (true) {
                    console.printf("Has the database been backed up? [Y/n] ");
                    String input = console.readLine();
                    if ("Y".equals(input)) {
                        result = true;
                        break;
                    } else if ("n".equals(input)) {
                        System.err.println("The database update can only be reverted by restoring a backup.\n"
                                           + "Back up the database before re-running this command again.");
                        result = false;
                        break;
                    }
                }
            }
        }
        return result;
    }
}
