/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.hl7.impl;

import ca.uhn.hl7v2.model.Message;
import org.junit.Test;
import org.mockito.Mockito;
import org.openvpms.hl7.io.Connector;
import org.openvpms.hl7.patient.PatientContext;
import org.openvpms.hl7.patient.PatientInformationService;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

/**
 * Tests the {@link PatientInformationServiceImpl}.
 *
 * @author Tim Anderson
 */
public class PatientInformationServiceImplTestCase extends AbstractServiceTest {

    /**
     * The admission service.
     */
    private PatientInformationService service;

    /**
     * Sets up the test case.
     */
    @Override
    public void setUp() {
        super.setUp();
        service = new PatientInformationServiceImpl(getArchetypeService(), getLookupService(), getEventServices(),
                                                    getDispatcher());

        PatientContext context = getContext();
        Mockito.when(context.getPatientId()).thenReturn(1001L);
        Mockito.when(context.getClinicianId()).thenReturn(2001L);
        Mockito.when(context.getCustomerId()).thenReturn(3001L);
    }

    /**
     * Tests the {@link PatientInformationService#admitted} method.
     *
     * @throws Exception for any error
     */
    @Test
    public void testAdmitted() throws Exception {
        String expected = "MSH|^~\\&|VPMS|Main Clinic|Cubex|Cubex|20140825085900||ADT^A01^ADT_A01|1200022|P|2.5||||||UTF-8\r" +
                          "EVN|A01|20140825085900|||||Main Clinic\r" +
                          "PID|1|1001|||Bar^Fido||20140701000000|M|||123 Broadwater Avenue^^Cape Woolamai^VIC^3058||(03) 12345678|(03) 98765432||||3001|||||||||||||||||CANINE^Canine^OpenVPMS|KELPIE^Kelpie^OpenVPMS\r" +
                          "PV1|1|U|^^^Main Clinic||||||||||||||2001^Blogs^Joe||3001|||||||||||||||||||||||||20140825085500\r" +
                          "OBX|1|NM|3141-9^BODY WEIGHT MEASURED^LN||10|kg^kilogram||||||||20140825085700\r" +
                          "AL1|1|MA|^Penicillin|U|Respiratory distress\r" +
                          "AL1|2|MA|^Pollen|U|Produces hives\r";

        service.admitted(getContext());
        assertTrue(getDispatcher().waitForMessage());
        checkMessage(expected);
    }

    /**
     * Tests the {@link PatientInformationService#admissionCancelled} method.
     *
     * @throws Exception for any error
     */
    @Test
    public void testAdmissionCancelled() throws Exception {
        String expected = "MSH|^~\\&|VPMS|Main Clinic|Cubex|Cubex|20140825085900||ADT^A11^ADT_A09|1200022|P|2.5||||||UTF-8\r" +
                          "EVN|A11|20140825085900|||||Main Clinic\r" +
                          "PID|1|1001|||Bar^Fido||20140701000000|M|||123 Broadwater Avenue^^Cape Woolamai^VIC^3058||(03) 12345678|(03) 98765432||||3001|||||||||||||||||CANINE^Canine^OpenVPMS|KELPIE^Kelpie^OpenVPMS\r" +
                          "PV1|1|U|^^^Main Clinic||||||||||||||2001^Blogs^Joe||3001|||||||||||||||||||||||||20140825085500\r" +
                          "OBX|1|NM|3141-9^BODY WEIGHT MEASURED^LN||10|kg^kilogram||||||||20140825085700\r";

        service.admissionCancelled(getContext());
        assertTrue(getDispatcher().waitForMessage());
        checkMessage(expected);
    }

    /**
     * Tests the {@link PatientInformationService#discharged} method.
     *
     * @throws Exception for any error
     */
    @Test
    public void testDischarged() throws Exception {
        String expected = "MSH|^~\\&|VPMS|Main Clinic|Cubex|Cubex|20140825085900||ADT^A03^ADT_A03|1200022|P|2.5||||||UTF-8\r" +
                          "EVN|A03|20140825085900|||||Main Clinic\r" +
                          "PID|1|1001|||Bar^Fido||20140701000000|M|||123 Broadwater Avenue^^Cape Woolamai^VIC^3058||(03) 12345678|(03) 98765432||||3001|||||||||||||||||CANINE^Canine^OpenVPMS|KELPIE^Kelpie^OpenVPMS\r" +
                          "PV1|1|U|^^^Main Clinic||||||||||||||2001^Blogs^Joe||3001|||||||||||||||||||||||||20140825085500\r" +
                          "AL1|1|MA|^Penicillin|U|Respiratory distress\r" +
                          "AL1|2|MA|^Pollen|U|Produces hives\r" +
                          "OBX|1|NM|3141-9^BODY WEIGHT MEASURED^LN||10|kg^kilogram||||||||20140825085700\r";

        service.discharged(getContext());
        assertTrue(getDispatcher().waitForMessage());
        checkMessage(expected);
    }

    /**
     * Tests the {@link PatientInformationService#updated} method.
     *
     * @throws Exception for any error
     */
    @Test
    public void testUpdated() throws Exception {
        String expected = "MSH|^~\\&|VPMS|Main Clinic|Cubex|Cubex|20140825085900||ADT^A08^ADT_A01|1200022|P|2.5||||||UTF-8\r" +
                          "EVN|A08|20140825085900|||||Main Clinic\r" +
                          "PID|1|1001|||Bar^Fido||20140701000000|M|||123 Broadwater Avenue^^Cape Woolamai^VIC^3058||(03) 12345678|(03) 98765432||||3001|||||||||||||||||CANINE^Canine^OpenVPMS|KELPIE^Kelpie^OpenVPMS\r" +
                          "PV1|1|U|^^^Main Clinic||||||||||||||2001^Blogs^Joe||3001|||||||||||||||||||||||||20140825085500\r" +
                          "OBX|1|NM|3141-9^BODY WEIGHT MEASURED^LN||10|kg^kilogram||||||||20140825085700\r" +
                          "AL1|1|MA|^Penicillin|U|Respiratory distress\r" +
                          "AL1|2|MA|^Pollen|U|Produces hives\r";

        service.updated(getContext());
        assertTrue(getDispatcher().waitForMessage());
        checkMessage(expected);
    }

    /**
     * Verifies that when {@link HL7Mapping#sendADT()} is {@code false}, no ADT messages are sent.
     */
    @Test
    public void testDisableADT() {
        service = new PatientInformationServiceImpl(getArchetypeService(), getLookupService(), getEventServices(),
                                                    getDispatcher()) {
            @Override
            protected void queue(Message message, Connector connector, HL7Mapping config) {
                fail("No ADT Message should be queued");
            }
        };
        getSender().getMapping().setSendADT(false);
        PatientContext patient = getContext();
        service.admitted(patient);
        service.admissionCancelled(patient);
        service.updated(patient);
        service.discharged(patient);
    }
}
