/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2020 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.customer.charge;

import org.openvpms.archetype.rules.patient.InvestigationActStatus;
import org.openvpms.component.business.domain.im.act.Act;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.component.im.util.IMObjectHelper;
import org.openvpms.web.resource.i18n.Messages;
import org.openvpms.web.workspace.workflow.investigation.OrderConfirmationDialog;
import org.openvpms.web.workspace.workflow.investigation.OrderConfirmationManager;

import java.util.ArrayList;
import java.util.List;

/**
 * Manages confirmation of interactive laboratory orders created during invoicing.
 *
 * @author Tim Anderson
 */
class ChargeOrderConfirmationManager {

    /**
     * The charge editor.
     */
    private final CustomerChargeActEditor editor;

    /**
     * The layout context.
     */
    private final LayoutContext context;


    /**
     * Constructs a {@link ChargeOrderConfirmationManager}.
     *
     * @param editor  the charge editor
     * @param context the layout context
     */
    ChargeOrderConfirmationManager(CustomerChargeActEditor editor, LayoutContext context) {
        this.editor = editor;
        this.context = context;
    }

    /**
     * Processes all investigations that need order confirmation.
     *
     * @param queue the queue used to queue confirmation dialogs.
     */
    public void process(EditorQueue queue) {
        List<Act> acts = new ArrayList<>();
        ChargeItemRelationshipCollectionEditor items = editor.getItems();
        InvestigationManager investigations = items.getEditContext().getInvestigations();
        for (Act investigation : investigations.getInvestigations()) {
            investigation = IMObjectHelper.reload(investigation);
            if (investigation != null && !InvestigationActStatus.CANCELLED.equals(investigation.getStatus())
                && InvestigationActStatus.CONFIRM.equals(investigation.getStatus2())) {
                acts.add(investigation);
            }
        }
        if (!acts.isEmpty()) {
            OrderConfirmationManager manager = new OrderConfirmationManager();
            List<OrderConfirmationDialog> dialogs = new ArrayList<>();
            for (Act investigation : acts) {
                OrderConfirmationDialog dialog = manager.createConfirmationDialog(investigation,
                                                                                  context.getHelpContext());
                if (dialog != null) {
                    dialogs.add(dialog);
                }
            }
            int count = 1;
            for (OrderConfirmationDialog dialog : dialogs) {
                String title = Messages.format("customer.charge.laboratory.order.title", count, dialogs.size());
                dialog.setTitle(title);
                queue.queue(dialog);
                count++;
            }
        }
    }

}
