/*
 * Version: 1.0
 *
 * The contents of this file are subject to the OpenVPMS License Version
 * 1.0 (the 'License'); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.openvpms.org/license/
 *
 * Software distributed under the License is distributed on an 'AS IS' basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * Copyright 2019 (C) OpenVPMS Ltd. All Rights Reserved.
 */

package org.openvpms.web.workspace.workflow.worklist;

import nextapp.echo2.app.Button;
import nextapp.echo2.app.event.ActionEvent;
import org.openvpms.archetype.rules.workflow.TaskStatus;
import org.openvpms.component.business.domain.im.act.Act;
import org.openvpms.component.business.service.archetype.helper.ActBean;
import org.openvpms.web.component.app.Context;
import org.openvpms.web.component.im.archetype.Archetypes;
import org.openvpms.web.component.im.layout.DefaultLayoutContext;
import org.openvpms.web.component.im.layout.LayoutContext;
import org.openvpms.web.component.im.util.IMObjectHelper;
import org.openvpms.web.component.workflow.DefaultTaskListener;
import org.openvpms.web.component.workflow.TaskEvent;
import org.openvpms.web.echo.button.ButtonSet;
import org.openvpms.web.echo.event.ActionListener;
import org.openvpms.web.echo.factory.ButtonFactory;
import org.openvpms.web.echo.help.HelpContext;
import org.openvpms.web.workspace.workflow.LocalClinicianContext;
import org.openvpms.web.workspace.workflow.scheduling.ScheduleCRUDWindow;


/**
 * Task CRUD window.
 *
 * @author Tim Anderson
 */
public class TaskCRUDWindow extends ScheduleCRUDWindow {

    /**
     * Transfer button identifier.
     */
    private static final String TRANSFER_ID = "transfer";


    /**
     * Constructs a {@code TaskCRUDWindow}.
     *
     * @param context the context
     * @param help    the help context
     */
    public TaskCRUDWindow(Context context, HelpContext help) {
        super(Archetypes.create("act.customerTask", Act.class), TaskActions.INSTANCE, context, help);
    }

    /**
     * Sets the object.
     *
     * @param object the object. May be {@code null}
     */
    @Override
    public void setObject(Act object) {
        super.setObject(object);
        getContext().setTask(object); // make available to macros etc
    }

    /**
     * Creates and edits a new object.
     */
    @Override
    public void create() {
        if (getContext().getWorkList() != null) {
            super.create();
        }
    }

    /**
     * Lays out the buttons.
     *
     * @param buttons the button row
     */
    @Override
    protected void layoutButtons(ButtonSet buttons) {
        super.layoutButtons(buttons);

        Button transfer = ButtonFactory.create(TRANSFER_ID, new ActionListener() {
            public void onAction(ActionEvent event) {
                onTransfer();
            }
        });

        buttons.add(createConsultButton());
        buttons.add(createCheckOutButton());
        buttons.add(transfer);
        buttons.add(createOverTheCounterButton());
        buttons.add(createFlowSheetButton());
    }

    /**
     * Enables/disables the buttons that require an object to be selected.
     *
     * @param buttons the button set
     * @param enable  determines if buttons should be enabled
     */
    @Override
    protected void enableButtons(ButtonSet buttons, boolean enable) {
        super.enableButtons(buttons, enable);
        if (enable) {
            Act act = getObject();
            enable = getActions().canCheckoutOrConsult(act);
        }
        buttons.setEnabled(CONSULT_ID, enable);
        buttons.setEnabled(CHECKOUT_ID, enable);
        buttons.setEnabled(TRANSFER_ID, enable);
    }

    /**
     * Creates a layout context for editing an object.
     *
     * @param help the help context
     * @return a new layout context
     */
    @Override
    protected LayoutContext createLayoutContext(HelpContext help) {
        // create a local context - don't want don't want to pick up the current clinician
        Context local = new LocalClinicianContext(getContext());
        return new DefaultLayoutContext(true, local, help);
    }

    /**
     * Transfers the selected task to a different worklist.
     */
    private void onTransfer() {
        final Act act = IMObjectHelper.reload(getObject());
        // make sure the act is still available
        if (act != null) {
            HelpContext help = getHelpContext().subtopic("transfer");
            TransferWorkflow transfer = new TransferWorkflow(act, getContext(), help);
            transfer.addTaskListener(new DefaultTaskListener() {
                public void taskEvent(TaskEvent event) {
                    onRefresh(act);
                }
            });
            transfer.start();
        }
    }

    protected static class TaskActions extends ScheduleActions {

        public static final TaskActions INSTANCE = new TaskActions();

        /**
         * Determines if a consultation or checkout can be performed on an act.
         *
         * @param act the act
         * @return {@code true} if consultation can be performed
         */
        @Override
        public boolean canCheckoutOrConsult(Act act) {
            boolean result = false;
            String status = act.getStatus();
            if (TaskStatus.PENDING.equals(status)
                    || TaskStatus.IN_PROGRESS.equals(status)
                || TaskStatus.BILLED.equals(status)) {
                ActBean bean = new ActBean(act);
                result = bean.getNodeParticipantRef("patient") != null;
            }
            return result;
        }
    }

}
